@echo OFF
@rem  Copyright (c) 2008-2010 by nexB, Inc. http://www.nexb.com/ - All rights reserved.
@rem  
@rem  Mostly inspired and derived from: 
@rem  http://cvs.zope.org/*checkout*/Zope/Attic/configure.bat
@rem  configure.bat,v 1.4 2003/06/25 00:27:49 
@rem  Licensed under the Zope Public License (ZPL) Version 2.0
@rem  see for details http://www.zope.org/Resources/License/ZPL-2.0 and
@rem  http://cvs.zope.org/*checkout*/Zope/Attic/LICENSE.txt?rev=HEAD&hideattic=0&content-type=text/plain

@rem  some ideas also borrowed from the ant.bat script from Apache Ant (Apache 2.0 license)
@rem  and from the Groovy.bat script from Groovy (Apache 1.1-style license)

@rem  To specify a Python interpreter to use instead of accepting the
@rem  result of this 'configure' script's autocheck, issue the command
@rem  "set PYTHON=\path\to\your\python" before running the script.

@rem  ===============================================
@rem  Borrowed from Groovy
@rem  Use explicit find.exe to prevent cygwin and others find.exe from being used
set FIND_EXE="find.exe"
if exist "%SystemRoot%\system32\find.exe" set FIND_EXE="%SystemRoot%\system32\find.exe"
if exist "%SystemRoot%\command\find.exe" set FIND_EXE="%SystemRoot%\command\find.exe"

@rem  cd to the folder of the dpl.bat file
cd %~dp0

set DPL_CMD_LINE_ARGS=

set PYTHON_VERSION=2.5

echo.

set FOUND_PYTHON=
goto getdplpython

:usage
 echo dpl [-d] [help] [command]
 echo.
 echo Run the DejaCode command line interpreter
 echo.
 echo   Options:
 echo     -d                  set the debug flag for advanced debugging display
 echo     help                print DejaCode help
 echo     command             DejaCode command to run
 echo.
 echo   Special:
 echo     To specify a Python interpreter to use, issue the command
 echo     "set PYTHON=drive:\path\to\your\python\executable" 
 echo     before running this script.
 echo.
 goto :EOF
 
:nopython
 echo.
 echo The DejaCode Toolkit requires Python $PYTHON_VERSION, which was not found at any usual location.
 echo.
 echo Please install Python by downloading version Python 2.5.4 from http://www.python.org
 echo.
 echo Alternatively, you can specify a Python installation location by setting 
 echo the environment variable PYTHON to the location of your already installed 
 echo Python version %PYTHON_VERSION% executable and restart DejaCode.
 echo.
 goto usage

:nosuchexecutable
 echo.
 echo ERROR: The Python interpreter ("%PYTHON%") set with the PYTHON environment 
 echo variable does not exist.
 echo.
 goto usage

:getdplpython
 @rem If in the DejaCode workbench, always use the Python shipped with it
 if not exist "..\..\python2.5\python.exe" goto getpythonfromenv
 set FOUND_PYTHON="..\..\python2.5\python.exe"
 goto rundpl

:getpythonfromenv
 @rem Finding a Python interpreter in the PYTHON envt var
 if "%PYTHON%"=="" goto findpythonfromreg
 if not exist "%PYTHON%" goto nosuchexecutable
 set FOUND_PYTHON="%PYTHON%"
 goto rundpl

:findpythonfromreg
 @rem Finding a Python interpreter in the global registry (installed for all users)
 regedit /E "%Temp%\python-for-dpl.reg" "HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\Python.exe"
 if not exist "%Temp%\python-for-dpl.reg" goto findpythonfromreg2
 for /F "tokens=1* delims==" %%A IN ('TYPE %Temp%\python-for-dpl.reg ^| %FIND_EXE% "Python.exe"') DO set FOUND_PYTHON=%%B
 set FOUND_PYTHON=%FOUND_PYTHON:"=%
 del %Temp%\python-for-dpl.reg
 if exist "%FOUND_PYTHON%" goto rundpl

:findpythonfromreg2
 @rem Finding a Python interpreter in the user registry (installed for the current user only)
 regedit /E "%Temp%\python-for-dpl2.reg" "HKEY_CURRENT_USER\SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\Python.exe"
 if not exist "%Temp%\python-for-dpl2.reg" goto findbasepython
 for /F "tokens=1* delims==" %%A IN ('TYPE %Temp%\python-for-dpl2.reg ^| %FIND_EXE% "Python.exe"') DO set FOUND_PYTHON=%%B
 set FOUND_PYTHON=%FOUND_PYTHON:"=%
 del %Temp%\python-for-dpl2.reg
 if exist "%FOUND_PYTHON%" goto rundpl

:findbasepython
 @rem Finding a Python interpreter at c:\Python25 the default install location
 if not exist "C:\python25\python.exe"  goto findbasepython2
 set FOUND_PYTHON="c:\Python25\python.exe"
 goto rundpl

:findbasepython2
 @rem Finding a Python interpreter at c:\Python2.5 another common install location
 if not exist "C:\python2.5\python.exe"  goto nopython
 set FOUND_PYTHON="c:\Python2.5\python.exe"
 goto rundpl
 
:rundpl
 @rem now that we have a python interpreter, run DPL
 @rem we check the version first
 "%FOUND_PYTHON%" -V 2>&1 | %FIND_EXE%  "%PYTHON_VERSION%" >NUL
 if errorlevel 1 goto nopython

 @rem Slurp the command line arguments. This loop allows for an unlimited number
 @rem of arguments (up to the command line limit, anyway). Borrowed from Apache ANT
:setupArgs
 if ""%1""=="""" goto nextInit
 set DPL_CMD_LINE_ARGS=%DPL_CMD_LINE_ARGS% %1
 shift
 goto setupArgs

:nextInit
"%FOUND_PYTHON%" -c "import sys, os; sys.path.remove(''); sys.path.insert(0, os.path.abspath('.')); from dejacode import interpreter; interpreter.main()" %DPL_CMD_LINE_ARGS%
:EOF
